import json
import numpy

from transiflow.utils import norm

from transiflow.Discretization import Discretization
from transiflow.CylindricalDiscretization import CylindricalDiscretization


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (numpy.int8, numpy.int16, numpy.int32, numpy.int64,
                            numpy.uint8, numpy.uint16, numpy.uint32, numpy.uint64)):

            return int(obj)
        elif isinstance(obj, (numpy.float16, numpy.float32, numpy.float64)):
            return float(obj)
        elif isinstance(obj, (complex, numpy.complex64, numpy.complex128)):
            return {'__complex__': True, 'real': obj.real, 'imag': obj.imag}
        elif isinstance(obj, numpy.ndarray):
            return obj.tolist()

        return json.JSONEncoder.default(self, obj)


def numpy_decoder(obj):
    if '__complex__' in obj:
        return complex(obj['real'], obj['imag'])
    return obj


class BaseInterface:
    '''This is the base class for all interfaces. We use this so we
    can write higher level methods such as pseudo-arclength
    continuation without knowing anything about the underlying methods
    such as the solvers that are present in the backend we are
    interfacing with.

    This class provides some basic implementations, which generally
    only work in serial, as well as some abstract methods that need to
    be implemented by the derived interface.

    '''

    def __init__(self, parameters, nx, ny, nz=1, dim=None, dof=None,
                 x=None, y=None, z=None, boundary_conditions=None):
        self.nx = nx
        self.ny = ny
        self.nz = nz

        if 'Problem Type' in parameters and 'Taylor-Couette' in parameters['Problem Type']:
            self.discretization = CylindricalDiscretization(parameters, nx, ny, nz, dim, dof,
                                                            x, y, z, boundary_conditions)
        else:
            self.discretization = Discretization(parameters, nx, ny, nz, dim, dof,
                                                 x, y, z, boundary_conditions)

        self.dim = dim or self.discretization.dim
        self.dof = dof or self.discretization.dof

        self.x = self.discretization.x
        self.y = self.discretization.y
        self.z = self.discretization.z

        self.parameters = parameters

        # Eigenvalue solver caching
        self._subspaces = None

    def _debug_print(self, *args):
        if self.parameters.get('Verbose', False):
            print('Debug:', *args, flush=True)

    def _debug_print_residual(self, string, jac, x, rhs):
        if self.parameters.get('Verbose', False):
            r = norm(jac @ x - rhs)
            self.debug_print(string, '{}'.format(r))

    def vector(self):
        '''Return a zero-initialized state vector suitable for the
        currently defined problem type.'''
        raise NotImplementedError()

    def vector_from_array(self, array):
        '''Create a state vector suitable for this interface from an
        array.'''
        return array

    def array_from_vector(self, vector):
        '''Create a numpy array suitable for postprocessing from a
        state vector.'''
        return vector

    def set_parameter(self, name, value):
        '''Set a parameter in the discretization.

        Parameters
        ----------
        name : str
            Name of the parameter in the parameter list.
        value : scalar
            Value of the parameter.

        '''
        self.discretization.set_parameter(name, value)

    def get_parameter(self, name):
        '''Get a parameter value from the discretization.

        Parameters
        ----------
        name : str
            Name of the parameter in the parameter list.

        '''
        return self.discretization.get_parameter(name)

    def save_json(self, name, obj):
        '''Save an object to a json file

        Parameters
        ----------
        name : str
            Name of the file.
        obj : Any
            Serializable object to save to the file.
        '''

        with open(name, 'w') as f:
            json.dump(obj, f, cls=NumpyEncoder)

    def save_parameters(self, name):
        '''Save the current parameter set to a file.

        Parameters
        ----------
        name : str
            Name of the file without extension.

        '''

        params_name = name + '.params'
        self.save_json(params_name, self.parameters)
        print('Wrote parameters to', params_name, flush=True)

    def save_state(self, name, x):
        '''Save the state ``x`` along with the current parameter set.

        Parameters
        ----------
        name : str
            Name of the file.
        x : array_like
            State at the current parameter values.

        '''
        self.save_parameters(name)

        if not name.endswith('.npy'):
            name += '.npy'

        numpy.save(name, x)
        print('Wrote state to', name, flush=True)

    def load_json(self, name):
        '''Load the an object from a json file.

        Parameters
        ----------
        name : str
            Name of the file without extension.

        '''
        with open(name, 'r') as f:
            return json.load(f, object_hook=numpy_decoder)

    def load_parameters(self, name):
        '''Load the parameter set from a file.

        Parameters
        ----------
        name : str
            Name of the file without extension.

        '''
        params_name = name + '.params'

        old_parameters = self.parameters.copy()
        self.parameters.update(self.load_json(params_name))
        print('Read parameters from', params_name, flush=True)

        for parameter_name, value in self.parameters.items():
            old_value = old_parameters.get(parameter_name)
            if value != old_value:
                print('Updated \'{}\' from {} to {}'.format(parameter_name, old_value, value),
                      flush=True)

    def load_state(self, name):
        '''Load the state ``x`` along with the current parameter set.

        Parameters
        ----------
        name : str
            Name of the file.

        '''
        self.load_parameters(name)

        if not name.endswith('.npy'):
            name += '.npy'

        x = numpy.load(name)
        print('Read state from', name, flush=True)

        return x

    def rhs(self, state):
        r'''Compute the right-hand side of the DAE. That is the
        right-hand side $F(u, p)$ in

        .. math:: M(p) \frac{\d u}{\d t} = F(u, p)

        Parameters
        ----------
        state : array_like
            State $u$ at which to evaluate $F(u, p)$

        Returns
        -------
        rhs : array_like
            The value of $F(u, p)$

        '''
        raise NotImplementedError()

    def jacobian(self, state):
        r'''Compute the Jacobian matrix $J(u, p)$ of the right-hand
        side of the DAE. That is the Jacobian matrix of $F(u, p)$ in

        .. math:: M(p) \frac{\d u}{\d t} = F(u, p)

        Parameters
        ----------
        state : array_like
            State $u$ at which to evaluate $J(u, p)$

        Returns
        -------
        jac : Matrix
            The matrix $J(u, p)$ in a suitable sparse format

        '''
        raise NotImplementedError()

    def mass_matrix(self):
        r'''Compute the mass matrix of the DAE. That is the mass
        matrix $M(p)$ in

        .. math:: M(p) \frac{\d u}{\d t} = F(u, p)

        Returns
        -------
        mass : Matrix
            The matrix $M(p)$ in a suitable sparse format

        '''
        raise NotImplementedError()

    def solve(self, jac, rhs, rhs2=None, V=None, W=None, C=None):
        '''Solve $J(u, p) y = x$ for $y$.

        Parameters
        ----------
        jac : Matrix
            The Jacobian matrix $J(u, p)$ as returned by
            ``jacobian()``.
        rhs : array_like
           The right-hand side vector $x$.
        rhs2 : array_like, optional
            Extension of ``rhs`` in case a bordered system is solved.
        V : array_like, optional
            Border case a bordered system ``[A, V; W^T, C] [y; y2] =
            [x; x2]`` is solved.
        W : array_like, optional
            Border case a bordered system ``[A, V; W^T, C] [y; y2] =
            [x; x2]`` is solved.
        C : matrix_like, optional
            Border case a bordered system ``[A, V; W^T, C] [y; y2] =
            [x; x2]`` is solved.

        '''
        raise NotImplementedError()

    def _eigs(self, jada_interface, jac_op, mass_op, prec_op, state,
              return_eigenvectors, enable_recycling):
        '''Internal helper for eigs()'''

        from jadapy import jdqz, orthogonalization

        parameters = self.parameters.get('Eigenvalue Solver', {})
        arithmetic = parameters.get('Arithmetic', 'complex')
        target = parameters.get('Target', 0.0)
        initial_subspace_dimension = parameters.get('Initial Subspace Dimension', 0)
        subspace_dimensions = [parameters.get('Minimum Subspace Dimension', 30),
                               parameters.get('Maximum Subspace Dimension', 60)]
        enable_recycling = parameters.get('Recycle Subspaces', enable_recycling)
        tol = parameters.get('Tolerance', 1e-7)
        num = parameters.get('Number of Eigenvalues', 5)

        if not enable_recycling:
            self._subspaces = None

        if not self._subspaces and initial_subspace_dimension > 0:
            # Use an inverse iteration to find guesses
            # for the eigenvectors closest to the target
            V = jada_interface.vector(initial_subspace_dimension)
            V[:, 0] = jada_interface.random()
            orthogonalization.normalize(V[:, 0])

            for i in range(1, initial_subspace_dimension):
                V[:, i] = jada_interface.prec(mass_op @ V[:, i-1])
                orthogonalization.orthonormalize(V[:, 0:i], V[:, i])

            self._subspaces = [V]
        elif self._subspaces:
            k = self._subspaces[0].shape[1]
            V = jada_interface.vector(k + 1)
            V[:, 0:k] = self._subspaces[0]
            V[:, k] = jada_interface.random()
            orthogonalization.orthonormalize(V[:, 0:k], V[:, k])

            gamma = numpy.sqrt(1 + abs(target) ** 2)
            W = jada_interface.vector(k + 1)
            W[:, 0:k] = self._subspaces[1]
            W[:, k] = (jac_op @ V[:, k]) * (1 / gamma) - (mass_op @ V[:, k]) * (target / gamma)
            orthogonalization.orthonormalize(W[:, 0:k], W[:, k])

            self._subspaces = [V, W]

        result = jdqz.jdqz(jac_op, mass_op, num, tol=tol, subspace_dimensions=subspace_dimensions, target=target,
                           interface=jada_interface, arithmetic=arithmetic, prec=prec_op,
                           return_eigenvectors=return_eigenvectors, return_subspaces=True,
                           initial_subspaces=self._subspaces)

        if return_eigenvectors:
            alpha, beta, v, q, z = result
            self._subspaces = [q, z]
            idx = range(len(alpha))
            idx = sorted(idx, key=lambda i: -(alpha[i] / beta[i]).real if (alpha[i] / beta[i]).real < 100 else 100)

            w = v.copy()
            eigs = alpha.copy()
            for i in range(len(idx)):
                w[:, i] = v[:, idx[i]]
                eigs[i] = alpha[idx[i]] / beta[idx[i]]
            return eigs[:num], w
        else:
            alpha, beta, q, z = result
            self._subspaces = [q, z]
            eigs = numpy.array(sorted(alpha / beta, key=lambda x: -x.real if x.real < 100 else 100))
            return eigs[:num]

    def eigs(self, state, return_eigenvectors=False, enable_recycling=False):
        r'''Compute the generalized eigenvalues of $\\beta * J(u, p) *
        v = \\alpha * M(p) * v$.

        Parameters
        ----------
        state : array_like
            State $u$ at which to evaluate $J(u, p)$.
        return_eigenvectors : bool, optional
            Whether to return the eigenvectors $v$.
        enable_recycling : bool, optional
            Whether to use the previous eigenvalue space as initial
            search space.

        Returns
        -------
        eigs : array_like
            Eigenvalues $\\alpha / \\beta$.
        v : array_like
            Corresponding eigenvectors in case ``return_eigenvectors``
            is ``True``.

        '''
        raise NotImplementedError()



import pytest
import numpy
import os

from transiflow import Continuation
from transiflow import Discretization
from transiflow import utils


def read_matrix(fname, m):
    from PyTrilinos import Epetra

    A = Epetra.CrsMatrix(Epetra.Copy, m, 27)

    dirname = os.path.dirname(__file__)
    with open(os.path.join(dirname, fname), 'r') as f:
        for i in f.readlines():
            r, c, v = [j.strip() for j in i.strip().split(' ') if j]
            r = int(r) - 1
            c = int(c) - 1
            v = float(v)
            if A.MyGlobalRow(r):
                A[r, c] = v
        A.FillComplete()

    return A

def read_vector(fname, m):
    from transiflow.interface import HYMLS as HYMLSInterface

    vec = HYMLSInterface.Vector(m)

    dirname = os.path.dirname(__file__)
    with open(os.path.join(dirname, fname), 'r') as f:
        for i, v in enumerate(f.readlines()):
            lid = m.LID(i)
            if lid != -1:
                vec[lid] = float(v.strip())
    return vec

def write_vector(vec, fname):
    dirname = os.path.dirname(__file__)
    with open(os.path.join(dirname, fname), 'w') as f:
        for i in range(len(vec)):
            f.write('%.16e\n' % vec[i])

def read_value(fname):
    val = 0
    dirname = os.path.dirname(__file__)
    with open(os.path.join(dirname, fname), 'r') as f:
        for v in f.readlines():
            val = float(v.strip())
    return val

def write_value(val, fname):
    dirname = os.path.dirname(__file__)
    with open(os.path.join(dirname, fname), 'w') as f:
        f.write('%.16e\n' % val)

def extract_sorted_row(A, i):
    values, indices = A.ExtractGlobalRowCopy(i)
    idx = sorted(range(len(indices)), key=lambda i: indices[i])
    return [indices[i] for i in idx], [values[i] for i in idx]

def extract_sorted_local_row(A, i):
    indices = A.jcoA[A.begA[i]:A.begA[i+1]]
    values = A.coA[A.begA[i]:A.begA[i+1]]
    idx = sorted(range(len(indices)), key=lambda i: indices[i])
    return [indices[i] for i in idx], [values[i] for i in idx]

def test_ldc():
    try:
        from transiflow.interface import HYMLS as HYMLSInterface
    except ImportError:
        pytest.skip("HYMLS not found")

    nx = 4
    ny = nx
    nz = nx
    dof = 4
    parameters = {'Reynolds Number': 100}
    n = nx * ny * nz * dof

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i+1

    interface = HYMLSInterface.Interface(parameters, nx, ny, nz, dof=dof)
    state = interface.vector_from_array(state)

    A = interface.jacobian(state)
    rhs = interface.rhs(state)

    B = read_matrix('data/ldc_%sx%sx%s.txt' % (nx, ny, nz), interface.solve_map)
    rhs_B = read_vector('data/ldc_rhs_%sx%sx%s.txt' % (nx, ny, nz), interface.map)

    for i in range(n):
        lid = interface.solve_map.LID(i)
        if lid == -1:
            continue

        print(i, lid)

        indices_A, values_A = extract_sorted_row(A, i)
        indices_B, values_B = extract_sorted_row(B, i)

        print('Expected:')
        print(indices_B)
        print(values_B)

        print('Got:')
        print(indices_A)
        print(values_A)

        assert len(indices_A) == len(indices_B)
        for j in range(len(indices_A)):
            assert indices_A[j] == indices_B[j]
            assert values_A[j] == pytest.approx(values_B[j])

    for i in range(n):
        lid = interface.map.LID(i)
        if lid == -1:
            continue

        print(i, lid, rhs[lid], rhs_B[lid])

        assert rhs_B[lid] == pytest.approx(rhs[lid])

def test_ldc_stretched_file():
    try:
        from transiflow.interface import HYMLS as HYMLSInterface
    except ImportError:
        pytest.skip("HYMLS not found")

    nx = 4
    ny = nx
    nz = nx
    dof = 4
    parameters = {'Reynolds Number': 100, 'Grid Stretching': True}
    n = nx * ny * nz * dof

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i+1

    interface = HYMLSInterface.Interface(parameters, nx, ny, nz, dof=dof)
    state = interface.vector_from_array(state)

    A = interface.jacobian(state)
    rhs = interface.rhs(state)

    B = read_matrix('data/ldc_stretched_%sx%sx%s.txt' % (nx, ny, nz), interface.solve_map)
    rhs_B = read_vector('data/ldc_stretched_rhs_%sx%sx%s.txt' % (nx, ny, nz), interface.map)

    for i in range(n):
        lid = interface.solve_map.LID(i)
        if lid == -1:
            continue

        print(i, lid)

        indices_A, values_A = extract_sorted_row(A, i)
        indices_B, values_B = extract_sorted_row(B, i)

        print('Expected:')
        print(indices_B)
        print(values_B)

        print('Got:')
        print(indices_A)
        print(values_A)

        assert len(indices_A) == len(indices_B)
        for j in range(len(indices_A)):
            assert indices_A[j] == indices_B[j]
            assert values_A[j] == pytest.approx(values_B[j])

    for i in range(n):
        lid = interface.map.LID(i)
        if lid == -1:
            continue

        print(i, lid, rhs[lid], rhs_B[lid])

        assert rhs_B[lid] == pytest.approx(rhs[lid])

def test_ldc_stretched(nx=4):
    try:
        from transiflow.interface import HYMLS as HYMLSInterface
    except ImportError:
        pytest.skip("HYMLS not found")

    ny = nx
    nz = nx
    dof = 4
    parameters = {'Reynolds Number': 100, 'Grid Stretching': True}
    n = nx * ny * nz * dof

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i+1

    discretization = Discretization(parameters, nx, ny, nz, dof=dof)
    B = discretization.jacobian(state)
    rhs_B = discretization.rhs(state)

    interface = HYMLSInterface.Interface(parameters, nx, ny, nz, dof=dof)
    state = interface.vector_from_array(state)

    A = interface.jacobian(state)
    rhs = interface.rhs(state)

    for i in range(n):
        lid = interface.solve_map.LID(i)
        if lid == -1:
            continue

        print(i, lid)

        indices_A, values_A = extract_sorted_row(A, i)
        indices_B, values_B = extract_sorted_local_row(B, i)

        print('Expected:')
        print(indices_B)
        print(values_B)

        print('Got:')
        print(indices_A)
        print(values_A)

        assert len(indices_A) == len(indices_B)
        for j in range(len(indices_A)):
            assert indices_A[j] == indices_B[j]
            assert values_A[j] == pytest.approx(values_B[j])

    for i in range(n):
        lid = interface.map.LID(i)
        if lid == -1:
            continue

        print(i, lid, rhs[lid], rhs_B[lid])

        assert rhs_B[i] == pytest.approx(rhs[lid])

def test_ldc8_stretched():
    test_ldc_stretched(8)

def test_prec(nx=4, parameters=None):
    try:
        from transiflow.interface import HYMLS as HYMLSInterface
        from PyTrilinos import Epetra
    except ImportError:
        pytest.skip("HYMLS not found")

    ny = nx
    nz = nx
    dof = 4
    n = nx * ny * nz * dof

    if not parameters:
        parameters = {'Reynolds Number': 100, 'Preconditioner': {'Number of Levels': 0}}

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i+1

    interface = HYMLSInterface.Interface(parameters, nx, ny, nz, dof=dof)
    state = interface.vector_from_array(state)

    if interface.comm.NumProc() > 1:
        if interface.comm.MyPID() == 0:
            pytest.skip("The preconditioner won't be exactly the same on multiple cores")
        return

    interface.jacobian(state)
    rhs = interface.rhs(state)

    rhs_sol = HYMLSInterface.Vector(interface.solve_map)
    prec_rhs = HYMLSInterface.Vector(interface.solve_map)
    rhs_sol.Import(rhs, interface.solve_importer, Epetra.Insert)

    interface.preconditioner.Compute()
    interface.preconditioner.ApplyInverse(rhs_sol, prec_rhs)

    # write_vector(prec_rhs, 'data/ldc_prec_rhs_%sx%sx%s.txt' % (nx, ny, nz))

    prec_rhs_B = read_vector('data/ldc_prec_rhs_%sx%sx%s.txt' % (nx, ny, nz), interface.solve_map)

    for i in range(n):
        lid = interface.solve_map.LID(i)
        if lid == -1:
            continue

        print(i, lid, prec_rhs[lid], prec_rhs_B[lid])

        assert prec_rhs_B[lid] == pytest.approx(prec_rhs[lid])

def test_multilevel_prec():
    parameters = {'Reynolds Number': 100, 'Preconditioner': {'Number of Levels': 2, 'Separator Length': 4}}
    test_prec(8, parameters)

def test_prec_stretched(nx=4, parameters=None):
    try:
        from transiflow.interface import HYMLS as HYMLSInterface
        from PyTrilinos import Epetra
    except ImportError:
        pytest.skip("HYMLS not found")

    ny = nx
    nz = nx
    dof = 4
    n = nx * ny * nz * dof

    if not parameters:
        parameters = {'Reynolds Number': 100, 'Grid Stretching': True, 'Preconditioner': {'Number of Levels': 0}}

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i+1

    interface = HYMLSInterface.Interface(parameters, nx, ny, nz, dof=dof)
    state = interface.vector_from_array(state)

    if interface.comm.NumProc() > 1:
        if interface.comm.MyPID() == 0:
            pytest.skip("The preconditioner won't be exactly the same on multiple cores")
        return

    interface.jacobian(state)
    rhs = interface.rhs(state)

    rhs_sol = HYMLSInterface.Vector(interface.solve_map)
    prec_rhs = HYMLSInterface.Vector(interface.solve_map)
    rhs_sol.Import(rhs, interface.solve_importer, Epetra.Insert)

    interface.preconditioner.Compute()
    interface.preconditioner.ApplyInverse(rhs_sol, prec_rhs)

    # write_vector(prec_rhs, 'data/ldc_stretched_prec_rhs_%sx%sx%s.txt' % (nx, ny, nz))

    prec_rhs_B = read_vector('data/ldc_stretched_prec_rhs_%sx%sx%s.txt' % (nx, ny, nz), interface.solve_map)

    for i in range(n):
        lid = interface.solve_map.LID(i)
        if lid == -1:
            continue

        print(i, lid, prec_rhs[lid], prec_rhs_B[lid])

        assert prec_rhs_B[lid] == pytest.approx(prec_rhs[lid])

def test_multilevel_prec_stretched():
    parameters = {'Reynolds Number': 100, 'Grid Stretching': True,
                  'Preconditioner': {'Number of Levels': 2, 'Separator Length': 4}}
    test_prec_stretched(8, parameters)

def test_bordered_prec():
    try:
        from transiflow.interface import HYMLS as HYMLSInterface
        from PyTrilinos import Epetra
    except ImportError:
        pytest.skip("HYMLS not found")

    nx = 4
    ny = nx
    nz = nx
    dof = 4
    n = nx * ny * nz * dof

    parameters = {'Reynolds Number': 100, 'Preconditioner': {'Number of Levels': 0}}

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i+1

    V = numpy.zeros(n)
    for i in range(n):
        state[i] = i+2

    W = numpy.zeros(n)
    for i in range(n):
        state[i] = i+3

    C = 42

    rhs_2 = 4

    interface = HYMLSInterface.Interface(parameters, nx, ny, nz, dof=dof)
    state = interface.vector_from_array(state)
    V = interface.vector_from_array(V)
    W = interface.vector_from_array(W)

    interface.jacobian(state)
    rhs = interface.rhs(state)

    rhs_2_sol = Epetra.SerialDenseMatrix(1, 1)
    rhs_2_sol[0, 0] = rhs_2

    prec_rhs_2 = Epetra.SerialDenseMatrix(1, 1)

    V_sol = HYMLSInterface.Vector(interface.solve_map)
    V_sol.Import(V, interface.solve_importer, Epetra.Insert)

    W_sol = HYMLSInterface.Vector(interface.solve_map)
    W_sol.Import(W, interface.solve_importer, Epetra.Insert)

    C_sol = Epetra.SerialDenseMatrix(1, 1)
    C_sol[0, 0] = C

    rhs_sol = HYMLSInterface.Vector(interface.solve_map)
    prec_rhs = HYMLSInterface.Vector(interface.solve_map)
    rhs_sol.Import(rhs, interface.solve_importer, Epetra.Insert)

    interface.preconditioner.SetBorder(V_sol, W_sol, C_sol)
    interface.preconditioner.Compute()
    interface.preconditioner.ApplyInverse(rhs_sol, rhs_2_sol, prec_rhs, prec_rhs_2)

    # write_vector(prec_rhs, 'ldc_bordered_prec_rhs_%sx%sx%s.txt' % (nx, ny, nz))
    # write_value(prec_rhs_2[0, 0], 'ldc_bordered_prec_rhs_2_%sx%sx%s.txt' % (nx, ny, nz))

    prec_rhs_B = read_vector('data/ldc_bordered_prec_rhs_%sx%sx%s.txt' % (nx, ny, nz), interface.solve_map)
    prec_rhs_2_B = read_value('data/ldc_bordered_prec_rhs_2_%sx%sx%s.txt' % (nx, ny, nz))

    for i in range(n):
        lid = interface.solve_map.LID(i)
        if lid == -1:
            continue

        print(i, lid, prec_rhs[lid], prec_rhs_B[lid])

        assert prec_rhs_B[lid] == pytest.approx(prec_rhs[lid])

    assert prec_rhs_2_B == prec_rhs_2[0, 0]

def test_norm():
    try:
        from transiflow.interface import HYMLS as HYMLSInterface
    except ImportError:
        pytest.skip("HYMLS not found")

    nx = 4
    ny = nx
    nz = nx
    dof = 4
    parameters = {}
    n = nx * ny * nz * dof

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i+1

    interface = HYMLSInterface.Interface(parameters, nx, ny, nz, dof=dof)

    state_dist = HYMLSInterface.Vector.from_array(interface.map, state)
    assert utils.norm(state) == utils.norm(state_dist)

    state_dist = HYMLSInterface.Vector.from_array(interface.solve_map, state)
    assert utils.norm(state) == utils.norm(state_dist)

def test_HYMLS(nx=4):
    try:
        from transiflow.interface import HYMLS as HYMLSInterface
        from PyTrilinos import Teuchos
    except ImportError:
        pytest.skip("HYMLS not found")

    numpy.random.seed(1234)

    ny = nx
    nz = nx

    parameters = Teuchos.ParameterList()
    parameters.set('Reynolds Number', 0)

    interface = HYMLSInterface.Interface(parameters, nx, ny, nz)
    continuation = Continuation(interface)

    x0 = interface.vector()
    x0 = continuation.newton(x0)

    start = 0
    target = 100
    ds = 100
    x = continuation.continuation(x0, 'Reynolds Number', start, target, ds)[0]

    assert x.Norm2() > 0



import numpy

from scipy import sparse
from scipy.sparse import linalg

from transiflow.interface import BaseInterface


class Interface(BaseInterface):
    '''This class defines an interface to the SciPy backend for the
    discretization. This backend can be used for low-resolution serial
    runs as well as implementing new problems in the discretization.

    See :mod:`.Discretization` for the descriptions of the constructor
    arguments.

    Parameters
    ----------
    parameters : dict
        Key-value pairs that can be used to modify parameters in the
        discretization as well as the linear solver and eigenvalue
        solver.

    '''
    def __init__(self, parameters, nx, ny, nz=1, dim=None, dof=None,
                 x=None, y=None, z=None, boundary_conditions=None):
        super().__init__(parameters, nx, ny, nz, dim, dof, x, y, z, boundary_conditions)

        # Select one pressure node to fix
        self.pressure_row = self.dim

        # Solver caching
        self.border_scaling = 1e-3
        self._lu = None
        self._prec = None

    def vector(self):
        return numpy.zeros(self.nx * self.ny * self.nz * self.dof)

    def rhs(self, state):
        return self.discretization.rhs(state)

    def jacobian(self, state):
        jac = self.discretization.jacobian(state)
        return sparse.csr_matrix((jac.coA, jac.jcoA, jac.begA)).tocsc()

    def mass_matrix(self):
        mass = self.discretization.mass_matrix()
        return sparse.csr_matrix((mass.coA, mass.jcoA, mass.begA), (mass.n, mass.n)).tocsc()

    def compute_bordered_matrix(self, jac, V=None, W=None, C=None, fix_pressure_row=False):
        '''Helper to compute a bordered matrix of the form ``[A, V; W^T, C]``.'''
        def _get_value(V, i, j):
            if not hasattr(V, 'shape') or len(V.shape) < 1:
                return V

            if len(V.shape) < 2:
                return V[i]

            return V[i, j]

        if fix_pressure_row and V is not None:
            self._debug_print(
                'Fixing pressure at row %d of the Jacobian matrix and adding the border' %
                self.pressure_row)
        elif V is not None:
            self._debug_print('Adding the border to the Jacobian matrix')
        else:
            self._debug_print('Fixing pressure at row %d of the Jacobian matrix' %
                              self.pressure_row)

        border_size = 0
        dtype = jac.dtype
        if V is not None and len(V.shape) > 1:
            border_size = V.shape[1]
            dtype = V.dtype
        elif V is not None:
            border_size = 1
            dtype = V.dtype

        extra_border_space = jac.shape[0] * border_size * 2 + border_size * border_size

        if W is None:
            W = V

        if C is None and border_size:
            C = numpy.zeros((border_size, border_size), dtype=jac.data.dtype)

        coA = numpy.zeros(jac.indptr[-1] + extra_border_space + 1, dtype=dtype)
        icoA = numpy.zeros(jac.indptr[-1] + extra_border_space + 1, dtype=int)
        begA = numpy.zeros(len(jac.indptr) + border_size, dtype=int)

        idx = 0
        for i in range(jac.shape[0]):
            if fix_pressure_row and i == self.pressure_row:
                coA[idx] = -1.0
                icoA[idx] = i
                idx += 1
                begA[i + 1] = idx
                continue

            for j in range(jac.indptr[i], jac.indptr[i + 1]):
                if not fix_pressure_row or jac.indices[j] != self.pressure_row:
                    coA[idx] = jac.data[j]
                    icoA[idx] = jac.indices[j]
                    idx += 1

            for j in range(border_size):
                coA[idx] = self.border_scaling * _get_value(W, i, j)
                icoA[idx] = jac.shape[0] + j
                idx += 1

            begA[i + 1] = idx

        for i in range(border_size):
            for j in range(jac.shape[0]):
                coA[idx] = self.border_scaling * _get_value(V, j, i)
                icoA[idx] = j
                idx += 1

            for j in range(border_size):
                coA[idx] = self.border_scaling * self.border_scaling * _get_value(C, j, i)
                icoA[idx] = jac.shape[0] + j
                idx += 1

            begA[jac.shape[0] + 1 + i] = idx

        n = len(begA) - 1
        return sparse.csc_matrix((coA, icoA, begA), (n, n))

    def _compute_factorization(self, jac, V, W, C):
        '''Compute the LU factorization of the (bordered) jacobian.'''

        if V is not None and hasattr(jac, 'lu') and jac.lu is not None and jac.bordered_lu:
            return

        if V is None and hasattr(jac, 'lu') and jac.lu is not None and not jac.bordered_lu:
            return

        self._lu = None
        self._prec = None
        jac.lu = None

        fix_pressure_row = self.dof > self.dim and self.pressure_row is not None
        A = self.compute_bordered_matrix(jac, V, W, C, fix_pressure_row)

        self._debug_print(
            'Computing the sparse LU factorization of the %s Jacobian matrix' % (
                'bordered ' if V is not None else ''))

        jac.lu = linalg.splu(A)
        jac.bordered_lu = V is not None

        # Cache the factorization for use in the iterative solver
        self._lu = jac.lu
        self._prec = linalg.LinearOperator(jac.shape,
                                           matvec=self._lu.solve,
                                           dtype=jac.dtype)

        self._debug_print(
            'Done computing the sparse LU factorization of the %s Jacobian matrix' % (
                'bordered ' if V is not None else ''))

    def _compute_preconditioner(self, jac, A, V, _W, _C):
        '''Compute the ILU factorization of the (bordered) jacobian.'''

        if V is not None and hasattr(jac, 'lu') and jac.lu is not None and jac.bordered_lu:
            return

        if V is None and hasattr(jac, 'lu') and jac.lu is not None and not jac.bordered_lu:
            return

        self._lu = None
        self._prec = None
        jac.lu = None

        self._debug_print('Computing the sparse ILU factorization of the Jacobian matrix')

        parameters = self.parameters.get('Preconditioner', {})
        jac.lu = linalg.spilu(A,
                              drop_tol=parameters.get('Drop Tolerance', None),
                              fill_factor=parameters.get('Fill Factor', None),
                              drop_rule=parameters.get('Drop Rule', None))
        jac.bordered_lu = A.shape != jac.shape

        # Cache the factorization for use in the iterative solver
        self._lu = jac.lu
        self._prec = linalg.LinearOperator(A.shape,
                                           matvec=self._lu.solve,
                                           dtype=A.dtype)

        self._debug_print('Done computing the sparse ILU factorization of the Jacobian matrix')

    def _lu_solve(self, A, rhs):
        if A.lu.L.dtype != rhs.dtype and numpy.dtype(rhs.dtype.char.upper()) == rhs.dtype:
            x = rhs.copy()
            x.real = self._lu_solve(A, rhs.real)
            x.imag = self._lu_solve(A, rhs.imag)
        else:
            x = A.lu.solve(rhs)

        return x

    def direct_solve(self, jac, rhs, rhs2=None, V=None, W=None, C=None):
        '''Solve $J(u, p) y = x$ for $y$ using a direct solver. See
        the :meth:`solve` function.

        '''
        x = rhs.copy()

        # Fix one pressure node
        if self.dof > self.dim:
            if len(x.shape) < 2:
                x[self.pressure_row] = 0
            else:
                x[self.pressure_row, :] = 0

        # First try to use an iterative solver with the previous
        # direct solver as preconditioner
        if self._prec and jac.dtype == x.dtype and jac.dtype == self._prec.dtype and \
           self.parameters.get('Use Iterative Solver', False):
            out, info, _ = gmres(jac, x, 5, 1e-8, prec=self._prec)
            if info == 0:
                return out

        if rhs2 is not None:
            x = numpy.append(rhs, rhs2 * self.border_scaling)

        # Use a direct solver instead
        self._compute_factorization(jac, V, W, C)

        if jac.bordered_lu:
            self._debug_print('Solving a bordered linear system')

            y = self._lu_solve(jac, x)

            border_size = 1
            if hasattr(rhs2, 'shape') and len(rhs2.shape) > 0:
                border_size = rhs2.shape[0]

            y1 = y[:-border_size]
            y2 = y[-border_size:] * self.border_scaling

            if numpy.isscalar(rhs2):
                y2 = y2.item()

            self._debug_print_residual('Done solving a bordered linear system with residual',
                                       jac, y1, rhs - V * y2)

            return y1, y2

        self._debug_print('Solving a linear system')

        y = self._lu_solve(jac, x)

        self._debug_print_residual('Done solving a linear system with residual', jac, y, rhs)

        return y

    def iterative_solve(self, jac, rhs, rhs2=None, V=None, W=None, C=None):
        '''Solve $J(u, p) y = x$ for $y$ using an iterative solver.
        See the :meth:`solve` function.

        '''
        x = rhs.copy()

        # Fix one pressure node
        if self.dof > self.dim:
            if len(x.shape) < 2:
                x[self.pressure_row] = 0
            else:
                x[self.pressure_row, :] = 0

        if rhs2 is not None:
            x = numpy.append(rhs, rhs2 * self.border_scaling)

        fix_pressure_row = self.dof > self.dim and self.pressure_row is not None
        A = self.compute_bordered_matrix(jac, V, W, C, fix_pressure_row)

        self._compute_preconditioner(jac, A, V, W, C)

        self._debug_print('Solving a linear system')

        parameters = self.parameters.get('Iterative Solver', {})
        restart = parameters.get('Restart', 100)
        maxit = parameters.get('Maximum Iterations', 1000)
        tol = parameters.get('Convergence Tolerance', 1e-6)

        y, info, gmres_iterations = gmres(A, x, maxit, tol, restart=restart, prec=self._prec)
        if info != 0:
            Exception('GMRES did not converge')

        if jac.bordered_lu:
            border_size = 1
            if hasattr(rhs2, 'shape') and len(rhs2.shape) > 0:
                border_size = rhs2.shape[0]

            y1 = y[:-border_size]
            y2 = y[-border_size:] * self.border_scaling

            self._debug_print_residual('Done solving a bordered linear system in %d iterations with residual' %
                                       gmres_iterations, jac, y1, rhs - V * y2)

            return y1, y2

        self._debug_print_residual('Done solving a linear system in %d iterations with residual' %
                                   gmres_iterations, jac, y, rhs)

        return y

    def solve(self, jac, rhs, rhs2=None, V=None, W=None, C=None):
        if self.parameters.get('Use Iterative Solver', False):
            return self.iterative_solve(jac, rhs, rhs2, V, W, C)

        return self.direct_solve(jac, rhs, rhs2, V, W, C)

    def eigs(self, state, return_eigenvectors=False, enable_recycling=False):
        from transiflow.interface.JaDa import Op

        parameters = self.parameters.get('Eigenvalue Solver', {})
        arithmetic = parameters.get('Arithmetic', 'complex')

        jac_op = Op(self.jacobian(state))
        mass_op = Op(self.mass_matrix())
        prec = None

        if self.parameters.get('Bordered Solver', False):
            from transiflow.interface.JaDa import BorderedInterface as JaDaInterface
        else:
            from transiflow.interface.JaDa import Interface as JaDaInterface

        jada_interface = JaDaInterface(self, jac_op, mass_op, jac_op.shape[0], numpy.complex128)
        if arithmetic == 'real':
            jada_interface = JaDaInterface(self, jac_op, mass_op, jac_op.shape[0])

        if not self.parameters.get('Bordered Solver', False):
            prec = jada_interface.shifted_prec

        return self._eigs(jada_interface, jac_op, mass_op, prec, state, return_eigenvectors,
                          enable_recycling)


def gmres(A, b, maxit, tol, restart=None, prec=None):
    '''Helper for running GMRES with different SciPy versions.

    :meta private:

    '''
    iterations = 0

    def callback(_r):
        nonlocal iterations
        iterations += 1

    if restart is None:
        restart = min(maxit, 100)

    maxiter = (maxit - 1) // restart + 1

    try:
        y, info = linalg.gmres(A, b, restart=restart, maxiter=maxiter,
                               rtol=tol, atol=0, M=prec,
                               callback=callback, callback_type='pr_norm')
    except TypeError:
        # Compatibility with SciPy <= 1.11
        y, info = linalg.gmres(A, b, restart=restart, maxiter=maxiter,
                               tol=tol, atol=0, M=prec,
                               callback=callback, callback_type='pr_norm')

    return y, info, iterations



from PyTrilinos import Epetra
from PyTrilinos import Teuchos

import numpy
import sys
import os

from transiflow.interface.Epetra import Vector
from transiflow.interface.Epetra import Interface as EpetraInterface

import HYMLS

def _set_default_parameter(parameterlist, name, value):
    if name not in parameterlist:
        parameterlist[name] = value

    return parameterlist[name]

def _convert_parameters(parameters, teuchos_parameters=None):
    if isinstance(parameters, Teuchos.ParameterList):
        return Teuchos.ParameterList(parameters)

    if teuchos_parameters is None:
        teuchos_parameters = Teuchos.ParameterList()

    for i, j in parameters.items():
        if isinstance(j, dict):
            sublist = teuchos_parameters.sublist(i)
            _convert_parameters(j, sublist)
        else:
            try:
                teuchos_parameters.set(i, j)
            except Exception:
                pass

    return teuchos_parameters

class Interface(EpetraInterface):
    '''This class defines an interface to the HYMLS backend for the
    discretization. This backend can be used for parallel simulations.
    It uses the HYMLS preconditioner for solving linear systems.

    The HYMLS backend partitions the domain into Cartesian subdomains,
    while solving linear systems on skew Cartesian subdomains to deal
    with the C-grid discretization. The subdomains will be distributed
    over multiple processors if MPI is used to run the application.

    See :mod:`.Discretization` for the descriptions of the constructor
    arguments.

    Parameters
    ----------
    parameters : dict
        Key-value pairs that can be used to modify parameters in the
        discretization as well as the preconditioner, iterative solver
        and eigenvalue solver.

    '''

    def __init__(self, parameters, nx, ny, nz=1, dim=None, dof=None,
                 boundary_conditions=None, comm=None):
        EpetraInterface.__init__(self, parameters, nx, ny, nz, dim, dof,
                                 boundary_conditions, comm)

        # Disable HYMLS output from MPI ranks != 0
        HYMLS.Tools.InitializeIO(self.comm)

        # Do the same for Python output
        self._original_stdout = sys.stdout
        if self.comm.MyPID() != 0:
            self._debug_print('PID %d: Disable output to stdout' % self.comm.MyPID())
            sys.stdout = open(os.devnull, 'w')

        self.teuchos_parameters = self.get_teuchos_parameters()

        partitioner = HYMLS.SkewCartesianPartitioner(self.teuchos_parameters, self.comm)
        partitioner.Partition()

        self.solve_map = partitioner.Map()
        self.solve_importer = Epetra.Import(self.solve_map, self.map)

        self.initialize()

    def __del__(self):
        HYMLS.Tools.PrintTiming()
        HYMLS.Tools.PrintMemUsage()

        if self.comm.MyPID() != 0:
            sys.stdout.close()
            sys.stdout = self._original_stdout
            self._debug_print('PID %d: Re-enable output to stdout' % self.comm.MyPID())

    def get_teuchos_parameters(self):
        '''Get the parameter list in Teuchos.ParameterList format.
        This is a copy of the internal list to solve issues with the
        ``Output Stream`` parameter and the fact that the ``get()``
        method actually sets values in the parameter list.

        :meta private:

        '''
        teuchos_parameters = _convert_parameters(self.parameters)

        problem_parameters = teuchos_parameters.sublist('Problem')
        _set_default_parameter(problem_parameters, 'nx', self.nx_global)
        _set_default_parameter(problem_parameters, 'ny', self.ny_global)
        _set_default_parameter(problem_parameters, 'nz', self.nz_global)

        _set_default_parameter(problem_parameters, 'Dimension', self.dim)
        _set_default_parameter(problem_parameters, 'Degrees of Freedom', self.dof)
        _set_default_parameter(problem_parameters, 'Equations', 'Stokes-C')

        _set_default_parameter(problem_parameters, 'x-periodic', self.discretization.x_periodic)
        _set_default_parameter(problem_parameters, 'y-periodic', self.discretization.y_periodic)
        _set_default_parameter(problem_parameters, 'z-periodic', self.discretization.z_periodic)

        solver_parameters = teuchos_parameters.sublist('Solver')
        solver_parameters.set('Initial Vector', 'Zero')
        solver_parameters.set('Left or Right Preconditioning', 'Right')

        iterative_solver_parameters = solver_parameters.sublist('Iterative Solver')
        iterative_solver_parameters.set('Output Stream', 0)
        maxit = _set_default_parameter(iterative_solver_parameters, 'Maximum Iterations', 1000)
        maxsize = _set_default_parameter(iterative_solver_parameters, 'Num Blocks', 100)
        _set_default_parameter(iterative_solver_parameters, 'Maximum Restarts', maxit // maxsize)
        _set_default_parameter(iterative_solver_parameters, 'Flexible Gmres', False)
        _set_default_parameter(iterative_solver_parameters, 'Convergence Tolerance', 1e-8)
        _set_default_parameter(iterative_solver_parameters, 'Output Frequency', 1)
        _set_default_parameter(iterative_solver_parameters, 'Show Maximum Residual Norm Only', False)
        _set_default_parameter(iterative_solver_parameters, 'Implicit Residual Scaling', 'Norm of RHS')
        _set_default_parameter(iterative_solver_parameters, 'Explicit Residual Scaling', 'Norm of RHS')

        prec_parameters = teuchos_parameters.sublist('Preconditioner')
        prec_parameters.set('Partitioner', 'Skew Cartesian')
        _set_default_parameter(prec_parameters, 'Separator Length', min(8, self.nx_global))
        _set_default_parameter(prec_parameters, 'Coarsening Factor', 2)
        _set_default_parameter(prec_parameters, 'Number of Levels', 1)

        coarse_solver_parameters = prec_parameters.sublist('Coarse Solver')
        _set_default_parameter(coarse_solver_parameters, "amesos: solver type", "Amesos_Superludist")

        return teuchos_parameters

    def _unset_parameter(self, name, original_parameters):
        '''Set a parameter in self.parameters back to its original value. '''

        if name in original_parameters:
            self.set_parameter(name, original_parameters[name])
            return

        if name in self.parameters:
            if hasattr(self.parameters, 'remove'):
                self.parameters.remove(name)
            else:
                del self.parameters[name]

    def initialize(self):
        '''Initialize the Jacobian and the preconditioner, but make sure the
        nonlinear part is also nonzero so we can replace all values
        later, rather than insert them.

        :meta private:

        '''
        # Backup the original parameters and put model parameters to 1
        parameter_names = ['Reynolds Number', 'Rayleigh Number',
                           'Prandtl Number', 'Rossby Parameter']

        original_parameters = {}
        for i in parameter_names:
            if i in self.parameters:
                original_parameters[i] = self.parameters[i]

            self.set_parameter(i, numpy.random.random())

        # Generate a Jacobian with a random state
        x = Vector(self.map)
        x.Random()
        self.jacobian(x)

        self._compute_scaling()
        self._scale_jacobian()

        self.preconditioner = HYMLS.Preconditioner(self.jac, self.teuchos_parameters)
        self.preconditioner.Initialize()

        self.solver = HYMLS.BorderedSolver(self.jac, self.preconditioner, self.teuchos_parameters)

        self._unscale_jacobian()

        # Put back the original parameters
        for i in parameter_names:
            self._unset_parameter(i, original_parameters)

    def solve(self, jac, rhs, rhs2=None, V=None, W=None, C=None, solver=None):
        if solver is None:
            solver = self.solver

        rhs_sol = Vector(self.solve_map)
        rhs_sol.Import(rhs, self.solve_importer, Epetra.Insert)

        x_sol = Vector(rhs_sol)

        self._compute_scaling()

        if rhs2 is not None:
            rhs2_sol = Epetra.SerialDenseMatrix(1, 1)
            rhs2_sol[0, 0] = rhs2

            x2_sol = Epetra.SerialDenseMatrix(1, 1)

            V_sol = Vector(self.solve_map)
            V_sol.Import(V, self.solve_importer, Epetra.Insert)

            self._scale_rhs(V_sol)

            W_sol = Vector(self.solve_map)
            W_sol.Import(W, self.solve_importer, Epetra.Insert)

            self._unscale_lhs(W_sol)

            C_sol = Epetra.SerialDenseMatrix(1, 1)
            C_sol[0, 0] = C

            solver.SetBorder(V_sol, W_sol, C_sol)

        self._scale_jacobian()
        self._scale_rhs(rhs_sol)

        self.preconditioner.Compute()
        if rhs2 is not None:
            solver.ApplyInverse(rhs_sol, rhs2_sol, x_sol, x2_sol)

            x2 = x2_sol[0, 0]
        else:
            solver.ApplyInverse(rhs_sol, x_sol)

        solver.UnsetBorder()

        self._unscale_jacobian()
        self._unscale_lhs(x_sol)

        x = Vector(rhs)
        x.Export(x_sol, self.solve_importer, Epetra.Insert)

        if rhs2 is not None:
            return x, x2

        return x

    def eigs(self, state, return_eigenvectors=False, enable_recycling=False):
        parameters = self.parameters.get('Eigenvalue Solver', {})
        arithmetic = parameters.get('Arithmetic', 'complex')

        if arithmetic == 'complex':
            from jadapy import ComplexEpetraInterface as EpetraInterface
            from transiflow.interface.JaDaHYMLS import ComplexInterface as JaDaHYMLSInterface
        else:
            from jadapy import EpetraInterface
            from transiflow.interface.JaDaHYMLS import BorderedInterface as JaDaHYMLSInterface

        jac_op = EpetraInterface.CrsMatrix(self.jacobian(state))
        mass_op = EpetraInterface.CrsMatrix(self.mass_matrix())

        self._compute_scaling()
        self._scale_matrix(jac_op)
        self._scale_matrix(mass_op)
        self._scale_jacobian()

        if arithmetic == 'complex':
            jada_interface = JaDaHYMLSInterface(self)
            prec = jada_interface.prec
        else:
            jada_interface = JaDaHYMLSInterface(self, preconditioned_solve=True)
            prec = None

        ret = self._eigs(jada_interface, jac_op, mass_op, prec,
                         state, return_eigenvectors, enable_recycling)

        if return_eigenvectors:
            self._unscale_lhs(ret[1])

        self._unscale_jacobian()

        return ret



import numpy
import os

import tempfile
import subprocess

class CrsMatrix:
    '''Compressed sparse row matrix used for assembly of the internal
    matrices. Can also be used for debugging purposes'''
    def __init__(self, coA=None, jcoA=None, begA=None, compress=True, m=None, n=None):
        self.coA = coA
        self.jcoA = jcoA
        self.begA = begA

        self._tmp = None

        if compress and coA is not None:
            self.compress()

        self.lu = None
        self.bordered_lu = False

        self._m = m
        self._n = n

    @property
    def m(self):
        '''Row dimension of the matrix'''
        if self._m:
            return self._m

        return len(self.begA) - 1

    @property
    def n(self):
        '''Column dimension of the matrix'''
        if self._n:
            return self._n

        return self.m

    @property
    def shape(self):
        '''Shape of the matrix'''
        return (self.m, self.n)

    @property
    def dtype(self):
        '''Shape of the matrix'''
        return self.coA.dtype

    def compress(self):
        ''' Remove zeros and merge duplicate entries, which may occur in the case of periodic
        boundary conditions.'''
        idx = 0
        beg = self.begA[0]
        for i in range(len(self.begA) - 1):
            unique_indices, inverse_indices = numpy.unique(self.jcoA[beg:self.begA[i+1]], return_inverse=True)

            values = numpy.zeros(len(unique_indices), dtype=self.coA.dtype)
            for orig_idx, inverse_idx in enumerate(inverse_indices):
                values[inverse_idx] += self.coA[beg + orig_idx]

            for j in range(len(unique_indices)):
                if abs(values[j]) > 1e-14:
                    self.jcoA[idx] = unique_indices[j]
                    self.coA[idx] = values[j]
                    idx += 1

            beg = self.begA[i+1]
            self.begA[i+1] = idx

    def solve(self, rhs):
        '''Solve a system using the self.lu property'''
        if self.lu.L.dtype != rhs.dtype and numpy.dtype(rhs.dtype.char.upper()) == rhs.dtype:
            x = rhs.copy()
            x.real = self.solve(rhs.real)
            x.imag = self.solve(rhs.imag)
        else:
            x = self.lu.solve(rhs)

        return x

    def __add__(self, B):
        coA = numpy.zeros(self.begA[-1] + B.begA[-1], dtype=self.dtype)
        jcoA = numpy.zeros(self.begA[-1] + B.begA[-1], dtype=int)
        begA = numpy.zeros(len(self.begA), dtype=int)

        idx = 0
        for i in range(self.n):
            for j in range(self.begA[i], self.begA[i+1]):
                found = False
                for k in range(B.begA[i], B.begA[i+1]):
                    if self.jcoA[j] == B.jcoA[k]:
                        coA[idx] = self.coA[j] + B.coA[k]
                        jcoA[idx] = self.jcoA[j]
                        idx += 1
                        found = True
                        break

                if not found:
                    coA[idx] = self.coA[j]
                    jcoA[idx] = self.jcoA[j]
                    idx += 1

            for j in range(B.begA[i], B.begA[i+1]):
                found = False
                for k in range(self.begA[i], self.begA[i+1]):
                    if B.jcoA[j] == self.jcoA[k]:
                        found = True
                        break

                if not found:
                    coA[idx] = B.coA[j]
                    jcoA[idx] = B.jcoA[j]
                    idx += 1

            begA[i+1] = idx

        return CrsMatrix(coA, jcoA, begA, False)

    def __iadd__(self, B):
        if self.coA is not None:
            A = self + B
            self.coA = A.coA
            self.jcoA = A.jcoA
            self.begA = A.begA
            return self

        self._tmp = B
        return self

    def __neg__(self):
        return CrsMatrix(-self.coA[:self.begA[-1]], self.jcoA[:self.begA[-1]], self.begA, False)

    def __sub__(self, B):
        return self + -B

    def __isub__(self, B):
        self += -B
        return self

    def __mul__(self, x):
        A = CrsMatrix(self.coA[:self.begA[-1]].copy(), self.jcoA[:self.begA[-1]].copy(),
                      self.begA.copy(), False)

        for i in range(self.n):
            for j in range(A.begA[i], A.begA[i+1]):
                A.coA[j] *= x
        return A

    def __truediv__(self, x):
        return self * (1 / x)

    def matvec(self, x):
        '''Return $y = A x$'''
        if len(x.shape) > 1:
            shape = list(x.shape)
            shape[0] = self.m
            b = numpy.zeros(shape, dtype=x.dtype)
            for i in range(self.m):
                for j in range(self.begA[i], self.begA[i+1]):
                    b[i, :] += self.coA[j] * x[self.jcoA[j], :]
            return b

        b = numpy.zeros(self.m, dtype=x.dtype)
        for i in range(self.m):
            for j in range(self.begA[i], self.begA[i+1]):
                b[i] += self.coA[j] * x[self.jcoA[j]]
        return b

    def __matmul__(self, x):
        return self.matvec(x)

    def __str__(self):
        out = ''
        for i in range(self.m):
            for j in range(self.begA[i], self.begA[i+1]):
                out += '%5d %5d %e\n' % (i, self.jcoA[j], self.coA[j])
        return out

    def to_coo(self):
        '''Convert the matrix to coordinate format.

        Returns
        -------
        coA : array_like
            Values
        icoA : array_like
            Row indices
        jcoA : array_like
            Column indices

        '''
        coA = numpy.zeros(self.begA[-1], dtype=self.dtype)
        icoA = numpy.zeros(self.begA[-1], dtype=int)
        jcoA = numpy.zeros(self.begA[-1], dtype=int)

        idx = 0
        for i in range(self.m):
            for j in range(self.begA[i], self.begA[i+1]):
                icoA[idx] = i
                jcoA[idx] = self.jcoA[j]
                coA[idx] = self.coA[j]
                idx += 1

        return coA, icoA, jcoA

    def to_dense(self):
        '''Convert the matrix to a dense matrix.

        Returns
        -------
        A : array_like
            Dense matrix

        '''
        A = numpy.zeros(self.shape, dtype=self.dtype)
        for i in range(self.m):
            for j in range(self.begA[i], self.begA[i+1]):
                A[i, self.jcoA[j]] = self.coA[j]

        return A

    def transpose(self):
        '''Return the transpose of the matrix.

        Returns
        -------
        B : CrsMatrix
            Transpose of the matrix

        '''
        coA, icoA, jcoA = self.to_coo()
        indices = sorted(range(self.begA[-1]), key=lambda i: jcoA[i])

        coB = numpy.zeros(self.begA[-1], dtype=self.dtype)
        jcoB = numpy.zeros(self.begA[-1], dtype=int)
        begB = numpy.zeros(len(self.begA), dtype=int)

        i = 0
        idx = 0
        for j in indices:
            while jcoA[j] > i:
                begB[i+1] = idx
                i += 1

            jcoB[idx] = icoA[j]
            coB[idx] = coA[j]
            idx += 1

        begB[i+1] = idx

        return CrsMatrix(coB, jcoB, begB, False)

    def dump(self, name):
        '''Dump the matrix to a file.

        Parmaters
        ---------
        name : string
            Name of the file

        '''
        with open(name, 'w') as f:
            out = '%%%%MatrixMarket matrix coordinate real general\n%d %d %d\n' % (self.m, self.n, self.begA[self.m])
            for i in range(self.m):
                for j in range(self.begA[i], self.begA[i+1]):
                    out += '%d %d %e\n' % (i+1, self.jcoA[j]+1, self.coA[j])
            f.write(out)

    def _get_index_list(self, idx, n):
        if isinstance(idx, int):
            idx = [idx]
        elif isinstance(idx, slice):
            start = idx.start
            if start is None:
                start = 0

            stop = idx.stop
            if stop is None:
                stop = n

            idx = numpy.arange(start, stop)

        return numpy.array(idx)

    def __getitem__(self, key):
        if not isinstance(key, tuple) or len(key) != 2:
            raise Exception('Key should be a tuple')

        iidx = self._get_index_list(key[0], self.m)
        jidx = self._get_index_list(key[1], self.n)

        if self.coA is None:
            # We must be using +=, which is implemented as
            # self.__setitem__(key, self.__getitem__(key).__iadd__(b))
            # so just return self here
            return CrsMatrix(m=len(iidx), n=len(jidx))

        coA = numpy.zeros(self.begA[-1], dtype=self.dtype)
        jcoA = numpy.zeros(self.begA[-1], dtype=int)
        begA = [0]

        index_list = -numpy.ones(self.n, dtype=int)
        for i, j in enumerate(jidx):
            index_list[j] = i

        idx = 0
        for i in iidx:
            for j in range(self.begA[i], self.begA[i+1]):
                if index_list[self.jcoA[j]] < 0:
                    continue

                coA[idx] = self.coA[j]
                jcoA[idx] = index_list[self.jcoA[j]]
                idx += 1

            begA.append(idx)

        if begA[-1] == 0:
            return 0

        if begA[-1] == 1:
            return coA[0]

        return CrsMatrix(coA[:idx], jcoA[:idx], numpy.array(begA),
                         compress=False, m=len(iidx), n=len(jidx))

    def assemble(self):
        '''Assemble the matrix after using the setter to assign
        values.'''
        assert self._tmp

        iidx = numpy.concatenate([i[0] for i in self._tmp])
        jidx = numpy.concatenate([i[1] for i in self._tmp])
        vals = numpy.concatenate([i[2] for i in self._tmp])

        sorted_idx = numpy.argsort(iidx)

        coA = numpy.zeros(len(iidx), dtype=vals[0].dtype)
        jcoA = numpy.zeros(len(iidx), dtype=int)
        begA = numpy.zeros(self.m + 1, dtype=int)

        idx = 0
        for i in range(self.m):
            for j in sorted_idx[idx:]:
                if iidx[j] > i:
                    break

                coA[idx] = vals[j]
                jcoA[idx] = jidx[j]
                idx += 1

            begA[i+1] = idx

        self._tmp = None

        self.coA = coA
        self.jcoA = jcoA
        self.begA = begA

        self.compress()

    def __setitem__(self, key, A):
        if not self._tmp:
            self._tmp = []

        iidx = self._get_index_list(key[0], self.m)
        jidx = self._get_index_list(key[1], self.n)

        if isinstance(A, CrsMatrix) and A._tmp is not None:
            self[key] = A._tmp
        elif isinstance(A, CrsMatrix):
            tmp = A.to_coo()
            self._tmp.append((iidx[tmp[1]], jidx[tmp[2]], tmp[0]))
        elif hasattr(A, 'shape'):
            tmp = list(numpy.where(abs(A) > 1e-14))
            self._tmp.append((iidx[tmp[0]], jidx[tmp[1]], A[tmp[0], tmp[1]]))
        else:
            self._tmp.append(([iidx[0]], [jidx[0]], [A]))

    def show(self, dof=None):
        '''Use vsm to show the structure of the matrix. vsm can
        currently be found in the mrilu directory of I-EMIC.'''
        def wrtbcsr(beg, jco, co, f):
            n = numpy.int32(len(beg) - 1)

            bc = numpy.int32(n.nbytes)
            os.write(f, bc.tobytes())
            os.write(f, n.tobytes())
            os.write(f, bc.tobytes())

            bc = numpy.int32(beg.nbytes)
            os.write(f, bc.tobytes())
            os.write(f, beg.tobytes())
            os.write(f, bc.tobytes())

            bc = numpy.int32(jco.nbytes)
            os.write(f, bc.tobytes())
            os.write(f, jco.tobytes())
            os.write(f, bc.tobytes())

            bc = numpy.int32(co.nbytes)
            os.write(f, bc.tobytes())
            os.write(f, co.tobytes())
            os.write(f, bc.tobytes())

        nnz = self.begA[-1]
        begA = numpy.ndarray(self.m + 1, numpy.int32)
        jcoA = numpy.ndarray(nnz, numpy.int32)
        coA = numpy.ndarray(nnz, float)

        begA[:] = self.begA
        jcoA[:] = self.jcoA[:nnz]
        coA[:] = self.coA[:nnz]

        if dof:
            idx = 0
            idx_map = [0] * self.m
            for d in range(dof):
                for i in range(d, self.m, dof):
                    idx_map[i] = idx
                    idx += 1

            idx = 0
            row_idx = 0
            for d in range(dof):
                for i in range(d, self.m, dof):
                    for j in range(self.begA[i], self.begA[i+1]):
                        jcoA[idx] = idx_map[self.jcoA[j]]
                        coA[idx] = self.coA[j]
                        idx += 1

                    row_idx += 1
                    begA[row_idx] = idx

        begA += 1
        jcoA += 1

        f, fname = tempfile.mkstemp()
        wrtbcsr(begA, jcoA, coA, f)
        os.close(f)

        subprocess.call(['vsm', fname])

        os.remove(fname)



import HYMLS

from PyTrilinos import Epetra

from jadapy import EpetraInterface
from jadapy import ComplexEpetraInterface

class PrecOp(EpetraInterface.Operator):
    def __init__(self, op, prec):
        super().__init__(op)
        self.prec = prec

    def ApplyInverse(self, x, y):
        self.prec.ApplyInverse(x, y)

        # Create a view here because this is an Epetra.MultiVector.
        z = EpetraInterface.Vector(Epetra.View, y, 0, y.NumVectors())
        z = self.op.proj(z)
        return y.Update(1.0, z, 0.0)

class Interface(EpetraInterface.EpetraInterface):

    def __init__(self, interface, *args, **kwargs):
        super().__init__(interface.solve_map)
        self.interface = interface
        self.parameters = interface.get_teuchos_parameters()
        self.parameters.setParameters(self.parameters.sublist('Eigenvalue Solver'))

        self.preconditioned_solve = kwargs.get('preconditioned_solve', False)

        self.interface.preconditioner.Compute()

    def solve(self, op, rhs, tol, maxit):
        solver_parameters = self.parameters.sublist('Solver')

        iterative_solver_parameters = solver_parameters.sublist('Iterative Solver')
        iterative_solver_parameters.set('Convergence Tolerance', tol if maxit > 1 else 1e-3)
        # iterative_solver_parameters.set('Maximum Iterations', maxit)

        if rhs.shape[1] == 2:
            solver_parameters.set('Complex', True)
        else:
            solver_parameters.set('Complex', False)

        out = EpetraInterface.Vector(rhs)

        epetra_op = EpetraInterface.Operator(op)
        if self.preconditioned_solve:
            epetra_precop = PrecOp(op, self.interface.preconditioner)
            solver = HYMLS.Solver(epetra_op, epetra_precop, self.parameters)
        else:
            solver = HYMLS.Solver(epetra_op, epetra_op, self.parameters)
        solver.ApplyInverse(rhs, out)

        return out

    def prec(self, x, *args):
        out = EpetraInterface.Vector(x)
        self.interface.preconditioner.ApplyInverse(x, out)
        return out

class ComplexPrecOp(EpetraInterface.Operator):
    def __init__(self, op, prec):
        super().__init__(op)
        self.prec = prec

    def ApplyInverse(self, x, y):
        self.prec.ApplyInverse(x, y)

        assert x.NumVectors() == 2
        # Create a view here because this is an Epetra.MultiVector.
        y = EpetraInterface.Vector(Epetra.View, y, 0, y.NumVectors())
        y = ComplexEpetraInterface.ComplexVector(y[:, 0], y[:, 1])

        z = self.op.proj(y)
        y *= 0.0
        y += z
        return 0

class ComplexInterface(ComplexEpetraInterface.ComplexEpetraInterface):

    def __init__(self, interface, *args, **kwargs):
        super().__init__(interface.solve_map)
        self.interface = interface
        self.parameters = interface.get_teuchos_parameters()
        self.parameters.setParameters(self.parameters.sublist('Eigenvalue Solver'))

        self.preconditioned_solve = kwargs.get('preconditioned_solve', False)

        self.interface.preconditioner.Compute()

    def solve(self, op, rhs, tol, maxit):
        solver_parameters = self.parameters.sublist('Solver')

        iterative_solver_parameters = solver_parameters.sublist('Iterative Solver')
        iterative_solver_parameters.set('Convergence Tolerance', tol if maxit > 1 else 1e-3)
        # iterative_solver_parameters.set('Maximum Iterations', maxit)

        solver_parameters.set('Complex', True)

        x = EpetraInterface.Vector(rhs.real.Map(), 2)
        y = EpetraInterface.Vector(rhs.real.Map(), 2)
        x[:, 0] = rhs.real
        x[:, 1] = rhs.imag

        epetra_op = ComplexEpetraInterface.Operator(op)
        if self.preconditioned_solve:
            epetra_precop = ComplexPrecOp(op, self.interface.preconditioner)
            solver = HYMLS.Solver(epetra_op, epetra_precop, self.parameters)
        else:
            solver = HYMLS.Solver(epetra_op, epetra_op, self.parameters)
        solver.ApplyInverse(x, y)

        out = ComplexEpetraInterface.ComplexVector(y[:, 0], y[:, 1])
        return out

    def prec(self, x, *args):
        y = EpetraInterface.Vector(x.real.Map(), 2)
        z = EpetraInterface.Vector(x.real.Map(), 2)
        y[:, 0] = x.real
        y[:, 1] = x.imag

        self.interface.preconditioner.ApplyInverse(y, z)

        out = ComplexEpetraInterface.ComplexVector(z[:, 0], z[:, 1])
        return out

class ShiftedOperator(object):
    def __init__(self, op):
        self.A = op.A
        self.B = op.B
        self.prec = op.prec
        self.Q = op.Q
        self.Z = op.Z
        self.Y = op.Y
        self.H = op.H
        self.alpha = op.alpha
        self.beta = op.beta

        self.dtype = self.Q.dtype
        self.shape = self.A.shape

    def matvec(self, x):
        return (self.A @ x) * self.beta - (self.B @ x) * self.alpha

class BorderedInterface(EpetraInterface.EpetraInterface):

    def __init__(self, interface, *args, **kwargs):
        super().__init__(interface.solve_map)
        self.interface = interface
        self.parameters = interface.get_teuchos_parameters()
        self.parameters.setParameters(self.parameters.sublist('Eigenvalue Solver'))

        self.preconditioned_solve = kwargs.get('preconditioned_solve', True)

        self.interface.preconditioner.Compute()

    def solve(self, op, rhs, tol, maxit):
        solver_parameters = self.parameters.sublist('Solver')

        iterative_solver_parameters = solver_parameters.sublist('Iterative Solver')
        iterative_solver_parameters.set('Convergence Tolerance', tol if maxit > 1 else 1e-3)
        # iterative_solver_parameters.set('Maximum Iterations', maxit)

        if rhs.shape[1] == 2:
            solver_parameters.set('Complex', True)
        else:
            solver_parameters.set('Complex', False)
        solver_parameters.set('Use Bordering', True)

        out = EpetraInterface.Vector(rhs)

        epetra_op = EpetraInterface.Operator(ShiftedOperator(op))
        if self.preconditioned_solve:
            solver = HYMLS.Solver(epetra_op, self.interface.preconditioner, self.parameters)
            solver.SetBorder(op.Z, op.Q)
            self.interface.preconditioner.Compute()
            solver.ApplyInverse(rhs, out)
            solver.UnsetBorder()
        else:
            raise Exception('Not implemented')

        return out

    def prec(self, x, *args):
        out = EpetraInterface.Vector(x)
        self.interface.preconditioner.ApplyInverse(x, out)
        return out

class ComplexBorderedInterface(ComplexEpetraInterface.ComplexEpetraInterface):

    def __init__(self, interface, *args, **kwargs):
        super().__init__(interface.solve_map)
        self.interface = interface
        self.parameters = interface.get_teuchos_parameters()
        self.parameters.setParameters(self.parameters.sublist('Eigenvalue Solver'))

        self.preconditioned_solve = kwargs.get('preconditioned_solve', True)

        self.interface.preconditioner.Compute()

    def solve(self, op, rhs, tol, maxit):
        solver_parameters = self.parameters.sublist('Solver')

        iterative_solver_parameters = solver_parameters.sublist('Iterative Solver')
        iterative_solver_parameters.set('Convergence Tolerance', tol if maxit > 1 else 1e-3)
        # iterative_solver_parameters.set('Maximum Iterations', maxit)

        solver_parameters.set('Complex', True)
        solver_parameters.set('Use Bordering', True)

        x = EpetraInterface.Vector(rhs.real.Map(), 2)
        y = EpetraInterface.Vector(rhs.real.Map(), 2)
        x[:, 0] = rhs.real
        x[:, 1] = rhs.imag

        m = op.Q.real.NumVectors()

        Q = EpetraInterface.Vector(rhs.real.Map(), m * 2)
        Q[:, 0:m] = op.Q.real
        Q[:, m:2*m] = op.Q.imag

        Z = EpetraInterface.Vector(rhs.real.Map(), m * 2)
        Z[:, 0:m] = op.Z.real
        Z[:, m:2*m] = op.Z.imag

        epetra_op = ComplexEpetraInterface.Operator(ShiftedOperator(op))
        if self.preconditioned_solve:
            solver = HYMLS.Solver(epetra_op, self.interface.preconditioner, self.parameters)
            solver.SetBorder(Z, Q)
            self.interface.preconditioner.Compute()
            solver.ApplyInverse(x, y)
            solver.UnsetBorder()
        else:
            raise Exception('Not implemented')

        out = ComplexEpetraInterface.ComplexVector(y[:, 0], y[:, 1])
        return out

    def prec(self, x, *args):
        y = EpetraInterface.Vector(x.real.Map(), 2)
        z = EpetraInterface.Vector(x.real.Map(), 2)
        y[:, 0] = x.real
        y[:, 1] = x.imag

        self.interface.preconditioner.ApplyInverse(y, z)

        out = ComplexEpetraInterface.ComplexVector(z[:, 0], z[:, 1])
        return out



import numpy
import pytest

from numpy.testing import assert_allclose

import matplotlib.pyplot as plt

from transiflow import utils

# Import common fixtures
from tests.jada_fixtures import * # noqa: F401, F403
from tests.jada_fixtures import check_eigenvalues

@pytest.fixture(autouse=True, scope='module')
def import_test():
    try:
        from transiflow.interface import JaDa # noqa: F401
    except ImportError:
        pytest.skip('jadapy not found')

class Operator:
    def __init__(self, A, B, alpha=0, beta=1):
        self.A = A
        self.B = B

        self.alpha = alpha
        self.beta = beta

        self.dtype = A.dtype
        self.shape = A.shape

    def matvec(self, x):
        return (self.A @ x) * self.beta - (self.B @ x) * self.alpha

    def proj(self, x):
        return x

def check_divfree(discretization, state):
    A = discretization.jacobian(state)
    x = A @ state
    for i in range(len(state)):
        if i % discretization.dof == discretization.dim:
            assert abs(x[i]) < 1e-14

def test_solve(interface, x, tol):
    from transiflow.interface import JaDa
    from jadapy.utils import norm

    numpy.random.seed(1234)

    jac_op = JaDa.Op(interface.jacobian(x))
    mass_op = JaDa.Op(interface.mass_matrix())
    jada_interface = JaDa.Interface(interface, jac_op, mass_op, len(x), preconditioned_solve=True)

    assert norm(x) > tol

    check_divfree(interface.discretization, x)

    b = numpy.zeros((x.shape[0], 1))
    b[:, 0] = x

    op = Operator(jac_op, mass_op)
    x = jada_interface.solve(op, b, tol, maxit=1)

    r = jac_op.matvec(x) - b
    r[interface.pressure_row] = 0

    assert norm(x) > tol
    assert norm(r) / norm(b) < tol

def test_shifted_solve(interface, x, tol):
    from transiflow.interface import JaDa
    from jadapy.utils import norm

    numpy.random.seed(1234)

    check_divfree(interface.discretization, x)
    assert norm(x) > tol

    jac_op = JaDa.Op(interface.jacobian(x))
    mass_op = JaDa.Op(interface.mass_matrix())

    b = numpy.zeros((x.shape[0], 1))
    b[:, 0] = x

    jada_interface = JaDa.Interface(interface, jac_op, mass_op, len(x), preconditioned_solve=True)

    op = Operator(jac_op, mass_op, 1, 2)
    x = jada_interface.solve(op, b, tol, maxit=1)

    r = (op.A @ x) * op.beta - (op.B @ x) * op.alpha - b
    r[interface.pressure_row] = 0

    assert norm(x) > tol
    assert norm(r) / norm(b) > tol

    jada_interface = JaDa.Interface(interface, jac_op, mass_op, len(x), preconditioned_solve=True, shifted=True)

    op = Operator(jac_op, mass_op, 1, 2)
    x = jada_interface.solve(op, b, tol, maxit=1)

    r = (op.A @ x) * op.beta - (op.B @ x) * op.alpha - b
    r[interface.pressure_row] = 0

    assert norm(x) > tol
    assert norm(r) / norm(b) < tol

def test_2D(arpack_eigs, interface, x, num_evs, tol, atol, interactive=False):
    from transiflow.interface import JaDa
    from jadapy import jdqz

    numpy.random.seed(1234)

    jac_op = JaDa.Op(interface.jacobian(x))
    mass_op = JaDa.Op(interface.mass_matrix())

    alpha, beta, q, z = jdqz.jdqz(jac_op, mass_op, num_evs, tol=tol, subspace_dimensions=[20, 40], return_subspaces=True)
    jdqz_eigs = numpy.array(sorted(alpha / beta, key=lambda x: abs(x)))
    jdqz_eigs = jdqz_eigs[:num_evs]

    assert_allclose(jdqz_eigs.real, arpack_eigs.real, rtol=0, atol=atol)
    assert_allclose(abs(jdqz_eigs.imag), abs(arpack_eigs.imag), rtol=0, atol=atol)

    H = utils.dot(q, q)
    assert_allclose(H, numpy.eye(q.shape[1]), rtol=0, atol=atol)

    if not interactive:
        return

    fig, ax = plt.subplots()
    ax.scatter(jdqz_eigs.real, jdqz_eigs.imag, marker='+')
    plt.show()

def test_complex_2D(arpack_eigs, interface, x, num_evs, tol, atol, interactive=False):
    from transiflow.interface import JaDa
    from jadapy import jdqz

    numpy.random.seed(1234)

    jac_op = JaDa.Op(interface.jacobian(x))
    mass_op = JaDa.Op(interface.mass_matrix())

    alpha, beta, v, q, z = jdqz.jdqz(jac_op, mass_op, num_evs, tol=tol, subspace_dimensions=[30, 60],
                                     arithmetic='complex', return_eigenvectors=True, return_subspaces=True)
    check_eigenvalues(jac_op, mass_op, alpha / beta, v, num_evs, atol)

    jdqz_eigs = numpy.array(sorted(alpha / beta, key=lambda x: abs(x)))

    assert_allclose(jdqz_eigs.real, arpack_eigs.real, rtol=0, atol=atol)
    assert_allclose(abs(jdqz_eigs.imag), abs(arpack_eigs.imag), rtol=0, atol=atol)

    H = utils.dot(q, q)
    assert_allclose(H, numpy.eye(q.shape[1]), rtol=0, atol=atol)

    if not interactive:
        return

    fig, ax = plt.subplots()
    ax.scatter(jdqz_eigs.real, jdqz_eigs.imag, marker='+')
    plt.show()

def test_prec_2D(arpack_eigs, interface, x, num_evs, tol, atol, interactive=False):
    from transiflow.interface import JaDa
    from jadapy import jdqz

    numpy.random.seed(1234)

    jac_op = JaDa.Op(interface.jacobian(x))
    mass_op = JaDa.Op(interface.mass_matrix())
    jada_interface = JaDa.Interface(interface, jac_op, mass_op, len(x))

    alpha, beta, v, q, z = jdqz.jdqz(jac_op, mass_op, num_evs, tol=tol, subspace_dimensions=[30, 60],
                                     return_eigenvectors=True, return_subspaces=True, prec=jada_interface.prec)

    jdqz_eigs = numpy.array(sorted(alpha / beta, key=lambda x: abs(x)))
    jdqz_eigs = jdqz_eigs[:num_evs]

    assert_allclose(jdqz_eigs.real, arpack_eigs.real, rtol=0, atol=atol)
    assert_allclose(abs(jdqz_eigs.imag), abs(arpack_eigs.imag), rtol=0, atol=atol)

    H = utils.dot(q, q)
    assert_allclose(H, numpy.eye(q.shape[1]), rtol=0, atol=atol)

    if not interactive:
        return

    fig, ax = plt.subplots()
    ax.scatter(jdqz_eigs.real, jdqz_eigs.imag, marker='+')
    plt.show()

def test_complex_prec_2D(arpack_eigs, interface, x, num_evs, tol, atol, interactive=False):
    from transiflow.interface import JaDa
    from jadapy import jdqz

    numpy.random.seed(1234)

    jac_op = JaDa.Op(interface.jacobian(x))
    mass_op = JaDa.Op(interface.mass_matrix())
    jada_interface = JaDa.Interface(interface, jac_op, mass_op, len(x), numpy.complex128)

    alpha, beta, v, q, z = jdqz.jdqz(jac_op, mass_op, num_evs, tol=tol, subspace_dimensions=[30, 60],
                                     arithmetic='complex', return_eigenvectors=True, return_subspaces=True,
                                     prec=jada_interface.prec)
    check_eigenvalues(jac_op, mass_op, alpha / beta, v, num_evs, atol)

    jdqz_eigs = numpy.array(sorted(alpha / beta, key=lambda x: abs(x)))

    assert_allclose(jdqz_eigs.real, arpack_eigs.real, rtol=0, atol=atol)
    assert_allclose(abs(jdqz_eigs.imag), abs(arpack_eigs.imag), rtol=0, atol=atol)

    H = utils.dot(q, q)
    assert_allclose(H, numpy.eye(q.shape[1]), rtol=0, atol=atol)

    if not interactive:
        return

    fig, ax = plt.subplots()
    ax.scatter(jdqz_eigs.real, jdqz_eigs.imag, marker='+')
    plt.show()

def test_complex_shifted_prec_2D(arpack_eigs, interface, x, num_evs, tol, atol, interactive=False):
    from transiflow.interface import JaDa
    from jadapy import jdqz

    numpy.random.seed(1234)

    jac_op = JaDa.Op(interface.jacobian(x))
    mass_op = JaDa.Op(interface.mass_matrix())
    jada_interface = JaDa.Interface(interface, jac_op, mass_op, len(x), numpy.complex128)

    alpha, beta, v, q, z = jdqz.jdqz(jac_op, mass_op, num_evs, tol=tol, subspace_dimensions=[30, 60],
                                     arithmetic='complex', return_eigenvectors=True, return_subspaces=True,
                                     prec=jada_interface.shifted_prec)
    check_eigenvalues(jac_op, mass_op, alpha / beta, v, num_evs, atol)

    jdqz_eigs = numpy.array(sorted(alpha / beta, key=lambda x: abs(x)))

    assert_allclose(jdqz_eigs.real, arpack_eigs.real, rtol=0, atol=atol)
    assert_allclose(abs(jdqz_eigs.imag), abs(arpack_eigs.imag), rtol=0, atol=atol)

    H = utils.dot(q, q)
    assert_allclose(H, numpy.eye(q.shape[1]), rtol=0, atol=atol)

    if not interactive:
        return

    fig, ax = plt.subplots()
    ax.scatter(jdqz_eigs.real, jdqz_eigs.imag, marker='+')
    plt.show()

def test_complex_solve_2D(arpack_eigs, interface, x, num_evs, tol, atol, interactive=False):
    from transiflow.interface import JaDa
    from jadapy import jdqz

    numpy.random.seed(1234)

    jac_op = JaDa.Op(interface.jacobian(x))
    mass_op = JaDa.Op(interface.mass_matrix())
    jada_interface = JaDa.Interface(interface, jac_op, mass_op, len(x), numpy.complex128)

    assert not jada_interface.preconditioned_solve
    assert not jada_interface.shifted

    alpha, beta, v, q, z = jdqz.jdqz(jac_op, mass_op, num_evs, tol=tol, subspace_dimensions=[30, 60],
                                     arithmetic='complex', return_eigenvectors=True, return_subspaces=True,
                                     interface=jada_interface)
    check_eigenvalues(jac_op, mass_op, alpha / beta, v, num_evs, atol)

    jdqz_eigs = numpy.array(sorted(alpha / beta, key=lambda x: abs(x)))

    assert_allclose(jdqz_eigs.real, arpack_eigs.real, rtol=0, atol=atol)
    assert_allclose(abs(jdqz_eigs.imag), abs(arpack_eigs.imag), rtol=0, atol=atol)

    H = utils.dot(q, q)
    assert_allclose(H, numpy.eye(q.shape[1]), rtol=0, atol=atol)

    if not interactive:
        return

    fig, ax = plt.subplots()
    ax.scatter(jdqz_eigs.real, jdqz_eigs.imag, marker='+')
    plt.show()

def test_complex_prec_solve_2D(arpack_eigs, interface, x, num_evs, tol, atol, interactive=False):
    from transiflow.interface import JaDa
    from jadapy import jdqz

    numpy.random.seed(1234)

    jac_op = JaDa.Op(interface.jacobian(x))
    mass_op = JaDa.Op(interface.mass_matrix())
    jada_interface = JaDa.Interface(interface, jac_op, mass_op, len(x), numpy.complex128,
                                    preconditioned_solve=True)

    assert jada_interface.preconditioned_solve
    assert not jada_interface.shifted

    alpha, beta, v, q, z = jdqz.jdqz(jac_op, mass_op, num_evs, tol=tol, subspace_dimensions=[30, 60],
                                     arithmetic='complex', return_eigenvectors=True, return_subspaces=True,
                                     interface=jada_interface)
    check_eigenvalues(jac_op, mass_op, alpha / beta, v, num_evs, atol)

    jdqz_eigs = numpy.array(sorted(alpha / beta, key=lambda x: abs(x)))

    assert_allclose(jdqz_eigs.real, arpack_eigs.real, rtol=0, atol=atol)
    assert_allclose(abs(jdqz_eigs.imag), abs(arpack_eigs.imag), rtol=0, atol=atol)

    H = utils.dot(q, q)
    assert_allclose(H, numpy.eye(q.shape[1]), rtol=0, atol=atol)

    if not interactive:
        return

    fig, ax = plt.subplots()
    ax.scatter(jdqz_eigs.real, jdqz_eigs.imag, marker='+')
    plt.show()

def test_complex_shifted_prec_solve_2D(arpack_eigs, interface, x, num_evs, tol, atol, interactive=False):
    from transiflow.interface import JaDa
    from jadapy import jdqz

    numpy.random.seed(1234)

    jac_op = JaDa.Op(interface.jacobian(x))
    mass_op = JaDa.Op(interface.mass_matrix())
    jada_interface = JaDa.Interface(interface, jac_op, mass_op, len(x), numpy.complex128,
                                    preconditioned_solve=True, shifted=True)

    assert jada_interface.preconditioned_solve
    assert jada_interface.shifted

    alpha, beta, v, q, z = jdqz.jdqz(jac_op, mass_op, num_evs, tol=tol, subspace_dimensions=[30, 60],
                                     arithmetic='complex', return_eigenvectors=True, return_subspaces=True,
                                     interface=jada_interface)
    check_eigenvalues(jac_op, mass_op, alpha / beta, v, num_evs, atol)

    jdqz_eigs = numpy.array(sorted(alpha / beta, key=lambda x: abs(x)))

    assert_allclose(jdqz_eigs.real, arpack_eigs.real, rtol=0, atol=atol)
    assert_allclose(abs(jdqz_eigs.imag), abs(arpack_eigs.imag), rtol=0, atol=atol)

    H = utils.dot(q, q)
    assert_allclose(H, numpy.eye(q.shape[1]), rtol=0, atol=atol)

    if not interactive:
        return

    fig, ax = plt.subplots()
    ax.scatter(jdqz_eigs.real, jdqz_eigs.imag, marker='+')
    plt.show()

def test_complex_shifted_prec_bordered_solve_2D(arpack_eigs, interface, x, num_evs, tol, atol, interactive=False):
    from transiflow.interface import JaDa
    from jadapy import jdqz

    numpy.random.seed(1234)

    jac_op = JaDa.Op(interface.jacobian(x))
    mass_op = JaDa.Op(interface.mass_matrix())
    jada_interface = JaDa.BorderedInterface(interface, jac_op, mass_op, len(x), numpy.complex128)

    alpha, beta, v, q, z = jdqz.jdqz(jac_op, mass_op, num_evs, tol=tol, subspace_dimensions=[30, 60],
                                     arithmetic='complex', return_eigenvectors=True, return_subspaces=True,
                                     interface=jada_interface)
    check_eigenvalues(jac_op, mass_op, alpha / beta, v, num_evs, atol)

    jdqz_eigs = numpy.array(sorted(alpha / beta, key=lambda x: abs(x)))

    assert_allclose(jdqz_eigs.real, arpack_eigs.real, rtol=0, atol=atol)
    assert_allclose(abs(jdqz_eigs.imag), abs(arpack_eigs.imag), rtol=0, atol=atol)

    H = utils.dot(q, q)
    assert_allclose(H, numpy.eye(q.shape[1]), rtol=0, atol=atol)
